IMPLEMENTATION MODULE Ausgleich; 

  (*------------------------------------------------------------------------*)
  (* Letzte Bearbeitung                                                     *)
  (*                                                                        *)
  (* 05.11.95 : Durchsicht                                                  *)
  (* 16.04.15 : In Prozedur PolyKoeff TrVek auf Zeiger umgestellt,          *)
  (*            Abfrage auf Laenge der Uebergabeparameter eingefuehrt.      *)
  (* 13.10.15, MRi: Ersetzen von LFLOAT(#) durch VAL(LONGREAL,#)            *)
  (*------------------------------------------------------------------------*)
  (* Licence : GNU Lesser General Public License (LGPL)                     *)
  (*------------------------------------------------------------------------*)

  (*
   * $Id: Ausgleich.mod,v 1.4 2016/04/15 15:42:15 mriedl Exp mriedl $
   *)

FROM SYSTEM     IMPORT TSIZE;
FROM Storage    IMPORT ALLOCATE,DEALLOCATE;
FROM Deklera    IMPORT MaxDim;
FROM Errors     IMPORT Fehler,Fehlerflag,ErrOut,FatalError;
FROM LongMath   IMPORT pi,sqrt,cos;
FROM SpezFunkt1 IMPORT FUNKTION,Clenshaw,Horner,RatPoly;
FROM MatLib     IMPORT SumVek,ENorm;
FROM LinLib     IMPORT HhLin;

PROCEDURE sqr(x : LONGREAL) : LONGREAL; BEGIN  RETURN x*x; END sqr;

PROCEDURE Ausgleichsgerade(VAR A,B   : LONGREAL;
                           VAR dA,dB : LONGREAL;
                           VAR Xq,Yq : LONGREAL;
                           VAR R     : LONGREAL;
                               dim   : CARDINAL;
                               T     : LONGREAL;
                           VAR X,Y   : ARRAY OF LONGREAL);

          CONST blas  = TRUE;
                eps   = 2.6E-15; (* Zahl, unterhalb derer eine Summe *)
                                (* als Null betrachtet wird.        *)
                MaxXY = 8191;

          VAR   i                           : CARDINAL;
                SumX,SumY,SumX2,SumY2,SumXY : LONGREAL;
                S,Sx,Sy,S1,S2,S3,S4,Dim     : LONGREAL;
                Nenner                      : LONGREAL;
                XY                          : POINTER TO
                                              ARRAY [0..MaxXY] OF LONGREAL;
                dimflag                     : BOOLEAN;

                (* Obergrenze in XY rein formal, vorsicht wenn Index"uber- *)
                (* pr"ufung des "Ubersetzer eingeschaltet. Siehe Abfrage   *)
                (* auf dimflag im Quellcode.                               *)

BEGIN
      Fehler:=FALSE;
      IF (dim <= 2) THEN
        ErrOut("Zuwenig Datenpunkte (Ausgleichsgerade)");
        Fehler:=TRUE;
        RETURN; 
      END;

      dimflag := (dim > (MaxXY+1)); (* Abfrage wegen Index"uberschreitung *)
      Dim:=VAL(LONGREAL,dim);

      IF NOT blas OR dimflag THEN (* Scheller, aber weniger genau ! *)
        SumX :=0.0; SumY :=0.0;
        SumX2:=0.0; SumY2:=0.0; SumXY:=0.0;
        FOR i:=0 TO dim-1 DO
          SumX  := SumX  + X[i];
          SumY  := SumY  + Y[i];
          SumX2 := SumX2 + X[i]*X[i];
          SumY2 := SumY2 + Y[i]*Y[i];
          SumXY := SumXY + X[i]*Y[i];
        END;
      ELSE
        SumX  := SumVek(X,1,dim);
        SumY  := SumVek(Y,1,dim);
        SumX2 := ENorm(X,1,dim); SumX2:=SumX2*SumX2;
        SumY2 := ENorm(Y,1,dim); SumY2:=SumY2*SumY2;
        ALLOCATE(XY,dim*TSIZE(LONGREAL));
        FOR i:=0 TO dim-1 DO (* Das ist auch noch nicht so doll ! *)
          XY^[i] := X[i]*Y[i];
        END;
        SumXY := SumVek(XY^,1,dim);
        DEALLOCATE(XY,dim*TSIZE(LONGREAL));
      END;
      Xq  := SumX / Dim;
      Yq  := SumY / Dim;

      Nenner  := SumX2 - (SumX*SumX / Dim);
      IF (ABS(Nenner) < (eps * SumXY)) OR (Nenner = 0.0) THEN 
        (* Nenner Null, wenn alle X_i = 1 oder 0           *)
        (* Dann sind die Eingabedaten aber nicht sinnvoll. *)
        A:=MAX(LONGREAL); dA:=MAX(LONGREAL);
        B:=MAX(LONGREAL); dB:=MAX(LONGREAL); R:=MAX(LONGREAL);
        Fehlerflag:=
        "Operation unbestimmt, Berechnung nicht moeglich (Ausgleichsgerade) !";
        Fehler:=TRUE; ErrOut(Fehlerflag);
        RETURN;
      END;
      A := (SumXY - Xq*SumY) / Nenner;
      B := Yq - A*Xq;

      S1:=0.0; S2:=0.0; S3:=0.0; S4:=0.0;
      FOR i:=0 TO dim-1 DO
        S1 := S1 + sqr(X[i] - Xq);
        S2 := S2 + sqr(Y[i] - Yq);
        S3 := S3 + (X[i] - Xq)*(Y[i] - Yq);
        S4 := S4 + sqr(A*(X[i]) + B - Y[i]);
      END;
      Sx := sqrt(S1);
      Sy := sqrt(S2);
      S  := sqrt(S4 / Dim);

      IF (Sx # 0.0) THEN 
        dA := T*S / (sqrt((Dim-2.0) / Dim)*Sx);
      ELSE
        dA := 0.0;
      END;
      dB := T*S /  sqrt (Dim-2.0);

      IF (Sx*Sy # 0.0) THEN
        R := S3 / (Sx*Sy);
      ELSE 
        R := 0.0;
      END;
END Ausgleichsgerade;

PROCEDURE TschebKoeff(    U,O    : LONGREAL;
                      VAR TKoeff : ARRAY OF LONGREAL;
                          Grad   : CARDINAL;
                          Funkt  : FUNKTION);

          (*--------------------------------------------------------*)
          (* C_i = {2 \oper N} \sum_{k=1}^N                         *)
          (*       f \left[                                         *)
          (*            \cos({\pi (k - 1/2) \over  N}               *)
          (*            \cos({\pi (j - 1) (k - 1/2) \over  N}       *)
          (*         \right]                                        *)
          (*--------------------------------------------------------*)

          VAR j,k                     : CARDINAL;
              y,N,OmU,OpU,Sum,Fac,Jm1 : LONGREAL;
              F                       : POINTER TO ARRAY [1..8192] OF LONGREAL;
BEGIN
      Fehler:=TRUE;
      IF (Grad+1 > HIGH(TKoeff)) THEN
        Fehlerflag:="Koeffizientenvektor zu klein (Ausgleich.TschebKoeff) !";
        Fehler:=TRUE; ErrOut(Fehlerflag);
        RETURN;
      END; 
      ALLOCATE(F,(Grad+1)*TSIZE(LONGREAL));
      IF (F = NIL) THEN
        Fehlerflag:="Kein Freispeicher vorhandden (Ausgleich.TschebKoeff) !";
        Fehler:=TRUE; ErrOut(Fehlerflag);
        RETURN;
      END;
      N := VAL(LONGREAL,Grad+1);
      OmU := 0.5*(O - U);
      OpU := 0.5*(O + U);
      FOR k:=1 TO Grad+1 DO
        y := cos(pi*(VAL(LONGREAL,k) - 0.5) / N);
        F^[k] := Funkt(y*OmU + OpU);
      END;
      Fac := 2.0 / N;
      FOR j:=1 TO Grad+1 DO
        Sum:=0.0; Jm1 := VAL(LONGREAL,j-1);
        FOR k:=1 TO Grad+1 DO
          Sum:=Sum + F^[k]*cos(pi*Jm1 * ((VAL(LONGREAL,k) - 0.5) / N));
        END;
        TKoeff[j-1]:=Fac*Sum;
      END;
      DEALLOCATE(F,(Grad+1)*TSIZE(LONGREAL));
END TschebKoeff;

PROCEDURE PolyKoeff(VAR PKoeff : ARRAY OF LONGREAL; (* Polynomkoeffizienten  *)
                    VAR TKoeff : ARRAY OF LONGREAL; (* Tschebytscheffkoeff.  *)
                        U,O    : LONGREAL;          (* Intervall f"ur Polynom *)
                        Grad   : CARDINAL);

          VAR N,i,j,k      : CARDINAL;
              TrVek        : POINTER TO ARRAY [0..8192] OF LONGREAL;
              Fac,Konst,Zw : LONGREAL;
BEGIN
      N := Grad + 1; (* Anzahl der Koeffizienten *)
      IF (N > HIGH(TKoeff)) OR (N > HIGH(PKoeff)) THEN
        Fehlerflag:="Koeffizientenvektoren zu klein (Ausgleich.PolyKoeff) !";
        Fehler:=TRUE; ErrOut(Fehlerflag);
        RETURN;
      END; 
      IF (N > 8191) THEN
        FatalError("Grad des Polynoms zu gro3 (Ausgleich.PolyKoeff) !");
      END;
      ALLOCATE(TrVek,N*TSIZE(LONGREAL));
      FOR i:=0 TO Grad DO PKoeff[i]:=0.0; TrVek^[i]:=0.0; END;
      PKoeff[0]:=TKoeff[Grad];
      FOR j:=N-1 TO 2 BY -1 DO
        FOR k:=N-j+1 TO 2 BY -1 DO
          Zw:=PKoeff[k-1];
          PKoeff[k-1]:= 2.0*PKoeff[k-2] - TrVek^[k-1];
          TrVek^[k-1]:=Zw;
        END;
        Zw:=PKoeff[0];
        PKoeff[0] := - TrVek^[0] + TKoeff[j-1];
        TrVek^[0]:=Zw;
      END; (* FOR j *)

      FOR j:=N TO 2 BY -1 DO PKoeff[j-1] := PKoeff[j-2] - TrVek^[j-1]; END;
      PKoeff[0] := - TrVek^[0] + 0.5*TKoeff[0];

      Konst := 2.0 / (O - U); (* Transfomiere auf Intervall [U,O] *)
      Fac := Konst;
      FOR j:=2 TO N DO
        PKoeff[j-1] := PKoeff[j-1] * Fac;
        Fac := Fac * Konst;
      END;
      Konst := 0.5*(U + O);
      FOR j:=1 TO N-1 DO
        FOR k:=N-1 TO j BY -1 DO
          PKoeff[k-1] := PKoeff[k-1] - Konst*PKoeff[k];
        END;
      END;
      DEALLOCATE(TrVek,N*TSIZE(LONGREAL));
END PolyKoeff;

PROCEDURE RatKoeff(    Funkt  : FUNKTION;
                       U,O    : LONGREAL;
                   VAR Ak,Bk  : ARRAY OF LONGREAL;
                       n,m    : CARDINAL);

          CONST  NDim = MaxDim DIV 2;

          VAR    i              : CARDINAL;
                 x,Delta,StdAbw : LONGREAL;
                 X,Y            : ARRAY [0..NDim-1] OF LONGREAL;

BEGIN
      x:=U; Delta := (O - U) / VAL(LONGREAL,NDim-2);
      FOR i:=0 TO NDim-1 DO X[i]:=x; Y[i]:=Funkt(x); x:=x + Delta; END;
      FitRat(X,Y,NDim,Ak,Bk,n,m,StdAbw);
END RatKoeff;

PROCEDURE FitTscheb(dim    : CARDINAL;
                    Grad   : CARDINAL;
                VAR X,Y    : ARRAY OF LONGREAL;
                VAR TKoeff : ARRAY OF LONGREAL;
                VAR U,O    : LONGREAL;
                VAR StdAbw : LONGREAL;
                VAR IFehl  : CARDINAL);

          (*------------------------------------------------------------*)
          (* Die Tschebytscheffkoeffizienten werden als L"osung der     *)
          (* des linearen Gleichungssystems F X - R = C erhalten mit :  *)
          (*                                                            *)
          (*    F_{i,0} = 1,                                            *)
          (*    F_{i,1} = x_j,                                          *)
          (*    F_{i,j} = 2 F_{i,j-1} X_j - T{i,j-2} X_j \forall j > 1  *)
          (*------------------------------------------------------------*)

          VAR  i,j,ij       : CARDINAL;
               Sum,Ri,K1,K2 : LONGREAL;
               F,C,X11      : POINTER TO ARRAY [0..8191] OF LONGREAL;

BEGIN
      IFehl:=0; Fehler:=FALSE;
      IF (dim < 2)  THEN IFehl:=1; ELSIF (dim > 8191) THEN IFehl:=2; END;
      IF (Grad < 1) THEN IFehl:=3; ELSIF (Grad > dim) THEN IFehl:=4; END;
      IF ((Grad+1)*dim > 8191) THEN IFehl:=5; END;
      CASE IFehl OF
        1 : Fehlerflag:='Anzahl der Datenpunkte kleiner zwei.';|
        2 : Fehlerflag:='Anzahl der Datenpunkte gr"o\3er als 8191.';|
        3 : Fehlerflag:='Grad des T-Polynoms kleiner eins.';|
        4 : Fehlerflag:='Grad des T-Polynoms gr"o\3er als Datenpunktanzahl.';|
        5 : Fehlerflag:='Fehlergleichungssystem zu gro\3.';
      ELSE
        Fehlerflag:='Kein Fehler';
      END;
      IF (IFehl # 0) THEN Fehler:=TRUE; ErrOut(Fehlerflag); RETURN; END;

      ALLOCATE(F,((Grad+1)*dim)*TSIZE(LONGREAL));
      ALLOCATE(C,dim*TSIZE(LONGREAL));
      ALLOCATE(X11,dim*TSIZE(LONGREAL));
      IF (F = NIL) OR (C = NIL) OR (X11 = NIL) THEN
        FatalError('Kein Freispeicher vorhanden (TschebFit) !');
      END;

      U:=X[0]; O:=X[0]; (* Transformiere X auf Intervall [-1,1] *)
      FOR i:=1 TO dim-1 DO
        IF (U > X[i]) THEN U := X[i]; END;
        IF (O < X[i]) THEN O := X[i]; END;
      END;
      K1 := 0.5*(O + U);
      K2 := 0.5*(O - U);
      FOR i:=0 TO dim-1 DO X11^[i] := (X[i] - K1) / K2; END;

      ij:=0;
      FOR i:=1 TO dim DO (* Berechne die Fehlergleichungen *)
        F^[ij] := 1.0; INC(ij);
        IF (Grad > 0) THEN F^[ij] := X11^[i-1]; INC(ij); END;
        FOR j:=2 TO Grad DO
          F^[ij]:= 2.0*X11^[i-1] * F^[ij-1] - F^[ij-2]; INC(ij);
        END;
      END;
      FOR i:=0 TO dim-1 DO C^[i]:= Y[i]; END;

      HhLin(F^,TKoeff,C^,dim,Grad+1);
      TKoeff[0]:=2.0*TKoeff[0];

      IF Fehler THEN
        Fehlerflag:='Keine L"osung gefunden (TschebKoeff) !';
        IFehl:=6; StdAbw:=MAX(LONGREAL);
      ELSE
        Sum:=0.0; StdAbw:=0.0;
        FOR i:=0 TO dim-1 DO
          Ri:= Y[i] - Clenshaw(X11^[i],Grad,TKoeff,-1.0,1.0);
          Sum:=Sum + Ri*Ri;
        END;
        IF (dim > Grad) THEN 
          StdAbw := sqrt(Sum / VAL(LONGREAL,dim - Grad)); 
        END;
      END;

      DEALLOCATE(X11,dim*TSIZE(LONGREAL));
      DEALLOCATE(C,dim*TSIZE(LONGREAL));
      DEALLOCATE(F,((Grad+1)*dim)*TSIZE(LONGREAL));
END FitTscheb;

PROCEDURE FitPoly(    dim     : CARDINAL;  (* Anzahl Datenpaare X,Y      *)
                  VAR X       : ARRAY OF LONGREAL;
                  VAR Y       : ARRAY OF LONGREAL;
                      Grad    : CARDINAL;  (* Grad des Polynoms          *)
                  VAR PKoeff  : ARRAY OF LONGREAL;(* Koeff. des Polynoms *)
                  VAR StdAbw  : LONGREAL;
                  VAR IFehl   : CARDINAL);

          VAR  i,j,ij      : CARDINAL;
               Sum,Ri,PotX : LONGREAL;
               F,C         : POINTER TO ARRAY [0..8192] OF LONGREAL;

BEGIN
      IFehl:=0; Fehler:=FALSE;
      IF (dim < 2)  THEN IFehl:=1; ELSIF (dim > 8192) THEN IFehl:=2; END;
      IF (Grad < 1) THEN IFehl:=3; ELSIF (Grad > dim) THEN IFehl:=4; END;
      IF ((Grad+1)*dim > 8192) THEN IFehl:=5; END;
      CASE IFehl OF
        1 : Fehlerflag:='Anzahl der Datenpunkte kleiner zwei.';|
        2 : Fehlerflag:='Anzahl der Datenpunkte gr"o3er als 8192.';|
        3 : Fehlerflag:='Grad des Polynoms kleiner eins.';|
        4 : Fehlerflag:='Grad des Polynoms gr"o3er als Datenpunktanzahl.';|
        5 : Fehlerflag:='Fehlergleichungssystem zu gro3.';
      ELSE
        Fehlerflag:='Kein Fehler';
      END;
      IF (IFehl # 0) THEN Fehler:=TRUE; ErrOut(Fehlerflag); RETURN; END;

      ALLOCATE(F,dim*(Grad+1)*TSIZE(LONGREAL));
      ALLOCATE(C,dim*TSIZE(LONGREAL));
      IF (F = NIL) OR (C = NIL) THEN
        FatalError('Kein Freispeicher vorhanden (FitPoly) !');
      END;

      ij:=0; (* Berechne "gepacktes" Fehlergleichungssystem *)
      FOR i:=0 TO dim-1 DO
        C^[i]:=Y[i]; PotX:=1.0;
        FOR j:=0 TO Grad DO F^[ij]:=PotX; PotX:=PotX * X[i]; INC(ij); END;
      END;

      HhLin(F^,PKoeff,C^,dim,Grad+1);

      IF Fehler THEN
        Fehlerflag:='Keine L"osung gefunden (FitPoly) !';
        IFehl:=6; StdAbw:=MAX(LONGREAL);
      ELSE
        Sum:=0.0; StdAbw:=0.0;
        FOR i:=0 TO dim-1 DO
          Ri:= Y[i] - Horner(X[i],PKoeff,Grad);
          Sum:=Sum + Ri*Ri;
        END;
        IF (dim > Grad) THEN 
          StdAbw := sqrt(Sum / VAL(LONGREAL,dim - Grad)); 
        END;
      END;

      DEALLOCATE(C,dim*TSIZE(LONGREAL));
      DEALLOCATE(F,dim*(Grad+1)*TSIZE(LONGREAL));
END FitPoly;

PROCEDURE FitRat(VAR X,Y    : ARRAY OF LONGREAL;
                     dim    : CARDINAL;
                 VAR Ak,Bk  : ARRAY OF LONGREAL;
                     n,m    : CARDINAL;
                 VAR StdAbw : LONGREAL);

          VAR MaxK,Sum,d      : LONGREAL;
              Grad,i,j,ij,max : CARDINAL;
              A,C,Koeff,PotX  : POINTER TO ARRAY [0..8191] OF LONGREAL;

BEGIN
      Grad := n + m + 2;
      IF (Grad > dim+1) THEN
        Fehler:=TRUE;
        Fehlerflag:='Zuwenige Datenpunkte f"ur die Interpolation (FitRat) !';
        ErrOut(Fehlerflag);
        RETURN;
      END;
      IF ((Grad)*dim > 8191) THEN
        FatalError('Fehlergleichungssystem zu gro\3 (FitRat) !');
      END;
      IF (n > m) THEN max:=n; ELSE max:=m; END;
      ALLOCATE(A,Grad*(dim+1)*TSIZE(LONGREAL));
      ALLOCATE(C,(dim+1)*TSIZE(LONGREAL));
      ALLOCATE(Koeff,Grad*TSIZE(LONGREAL));
      ALLOCATE(PotX,(max+1)*TSIZE(LONGREAL));
      IF (A = NIL) OR (C = NIL) OR (Koeff = NIL) OR (PotX = NIL) THEN
        FatalError('Kein Freispeicher vorhanden (FitRat) !');
      END;

      ij:=0; PotX^[0]:=1.0;
      FOR i:=0 TO dim-1 DO
        FOR j:=1 TO max DO PotX^[j] := X[i]*PotX^[j-1]; END;
        FOR j:=0 TO n DO A^[ij] :=        PotX^[j]; INC(ij); END;
        FOR j:=0 TO m DO A^[ij] := - Y[i]*PotX^[j]; INC(ij); END;
        C^[i]:=0.0;
      END; (* FOR i *)
      FOR i:=0 TO n+m+1 DO A^[ij]:=1.0; INC(ij); END;
      C^[dim]:=1.0;

      HhLin(A^,Koeff^,C^,(dim+1),Grad);

      IF (Koeff^[n+1] # 0.0) THEN
        MaxK := Koeff^[n+1];
      ELSE
        MaxK:=Koeff^[0];
        FOR i:=1 TO Grad-1 DO
          IF (ABS(Koeff^[i]) > ABS(MaxK)) THEN MaxK:=Koeff^[i]; END;
        END;
      END;
      FOR i:=0 TO Grad-1 DO Koeff^[i] := Koeff^[i] / MaxK; END;
      FOR i:=n TO HIGH(Ak) DO Ak[i] := 0.0; END;
      FOR i:=m TO HIGH(Bk) DO Bk[i] := 0.0; END;
      FOR i:=0 TO n DO Ak[i] := Koeff^[i]; END;
      FOR i:=0 TO m DO Bk[i] := Koeff^[i+n+1]; END;

      Sum:=0.0;
      FOR i:=0 TO dim-1 DO
        d := Y[i] - RatPoly(X[i],Ak,Bk,n,m);
        Sum:=Sum + d*d;
      END;
      StdAbw:=0.0;
      IF (dim > Grad) THEN 
        StdAbw := sqrt(Sum / VAL(LONGREAL,dim - Grad)); 
      END;

      DEALLOCATE(PotX,(max+1)*TSIZE(LONGREAL));
      DEALLOCATE(Koeff,Grad*TSIZE(LONGREAL));
      DEALLOCATE(C,(dim+1)*TSIZE(LONGREAL));
      DEALLOCATE(A,Grad*(dim+1)*TSIZE(LONGREAL));
END FitRat;

PROCEDURE Spline(VAR Xa,Ya : ARRAY OF LONGREAL; (* Eingabevektoren         *)
                     N     : CARDINAL;          (* Anzahl der Eingabewerte *)
                     X1,XM : LONGREAL;  
                 VAR Xn,Yn : ARRAY OF LONGREAL; (* Ausgabevektor           *)
                     M     : CARDINAL);         (* Anzahl Splinepunkte     *)

          (*------------------------------------------------------*)
          (* In Unterroutinen mit (**) gekenzeichnete Parameter   *)
          (* werden nur aus Effizienzgr"unden als VAR deklariert. *)
          (*------------------------------------------------------*)

          CONST     MaxVek = 64*1024;

  PROCEDURE SI(VAR X,Y,Z : ARRAY OF LONGREAL; (**)
               VAR V     : LONGREAL;          (**)
                   N     : CARDINAL) : LONGREAL;

            VAR i        : CARDINAL;
                zw,ai,hi : LONGREAL;
  BEGIN
        IF (V > X[0]) AND (V < X[N-1]) THEN
          i:=N-2;
          REPEAT DEC(i); zw:=V - X[i]; UNTIL (zw >= 0.0) OR (i = 1);
          hi:=X[i+1]-X[i];
          ai:=zw*(Z[i+1] - Z[i]) / (6.0*hi) + 0.5*Z[i];
          ai:=zw*ai + (Y[i+1]-Y[i]) / hi-hi*(2.0*Z[i]+Z[i+1])/6.0;
          RETURN zw*ai+Y[i];
        ELSIF (V = X[0]) THEN
          RETURN Y[0]
        ELSE
          RETURN Y[N-1];
        END;
  END SI;

  PROCEDURE spline(VAR X,Y  : ARRAY OF LONGREAL;   (**)
                       N    : CARDINAL;
                   VAR XInt : ARRAY OF LONGREAL;   (**)
                   VAR YInt : ARRAY OF LONGREAL;
                       M    : CARDINAL);

            (* F"uhrt die eigentliche Spline-Intepolation aus.       *)
            (* X,Y       : Vorgegebene Datenpunlte.                  *)
            (* N         : Anzahl der Datenpunkte.                   *)
            (* XInt,YInt : Splineinterpolierte Wertepaare.           *)
            (* M         : Anzahl der erzeugten Datenpunkte in den   *)
            (*             Feldern XInt,YInt.                        *) 

            VAR   i      : CARDINAL;
                  C,D,Z  : POINTER TO ARRAY [0..MaxVek] OF LONGREAL;
                  Factor : LONGREAL;
  BEGIN
        ALLOCATE(C,N*TSIZE(LONGREAL));
        ALLOCATE(D,N*TSIZE(LONGREAL));
        ALLOCATE(Z,N*TSIZE(LONGREAL));

        IF (C = NIL) OR (D = NIL) OR (Z = NIL) THEN
          FatalError("Zuwenig Freispeicher vorhanden (Ausgleich.Spline) !");
        END;
        D^[0]:=1.0; C^[0]:=0.5; Z^[0]:=0.5;
        FOR i:=1 TO N-2 DO
          D^[i]:=2.0*(X[i+1] - X[i-1]);
          C^[i]:=X[i+1] - X[i];
          Z^[i]:=6.0*((Y[i+1]-Y[i]) / (X[i+1]-X[i]) -
                     (Y[i]-Y[i-1]) / (X[i]-X[i-1]));
        END;
        D^[N-1]:=1.0; C^[N-2]:=0.0;
        C^[N-1]:=0.0; Z^[N-1]:=0.0;
        FOR i:=1 TO N-1 DO
          Factor := C^[i-1] / D^[i-1];
          D^[i]:=D^[i] - Factor*C^[i-1];
          Z^[i]:=Z^[i] - Factor*Z^[i-1];
        END;
        Z^[N-1]:=Z^[N-1] / D^[N-1];
        FOR i:=0 TO N-2 DO
          Z^[N-i-2]:=(Z^[N-i-2] - C^[N-i-2]*Z^[N-i-1]) / D^[N-i-2];
        END;
        FOR i:=0 TO M-1 DO YInt[i] := SI(X,Y,Z^,XInt[i],N); END;
        DEALLOCATE(Z,N*TSIZE(LONGREAL));
        DEALLOCATE(D,N*TSIZE(LONGREAL));
        DEALLOCATE(C,N*TSIZE(LONGREAL));
  END spline;

          VAR   i         : CARDINAL;
                x,DeltaX  : LONGREAL;
                Xalt,Yalt : POINTER TO ARRAY [0..MaxVek-1] OF LONGREAL; 
BEGIN
      IF (N > MaxVek+1) OR (M > MaxVek+1) THEN
        FatalError('M oder N gr"o3er MaxVek  (Spline)');
      END;
      IF (N >= 1) AND (M >= 1) THEN
        IF ((X1 >= Xa[0]) AND (XM <= Xa[N-1])) AND (M >= 1) THEN
          ALLOCATE(Xalt,N*TSIZE(LONGREAL));
          ALLOCATE(Yalt,N*TSIZE(LONGREAL));
          IF (Xalt = NIL) OR (Yalt = NIL) THEN
            FatalError("Zuwenig Freispeicher vorhanden (Ausgleich.Spline) !");
          END;
          FOR i:=0 TO N-1 DO Xalt^[i]:=Xa[i]; Yalt^[i]:=Ya[i]; END;
          DeltaX := ABS(XM - X1) / VAL(LONGREAL,M);
          x:=X1;
          FOR i:=0 TO M-1 DO Xn[i]:=x; x:=x + DeltaX; END;
          spline(Xalt^,Yalt^,N,Xn,Yn,M); (* Eigendliche Transformation. *)
          DEALLOCATE(Yalt,N*TSIZE(LONGREAL));
          DEALLOCATE(Xalt,N*TSIZE(LONGREAL));
        ELSE
          FatalError('Wert(e) au3erhalb des zul"assigen Bereichs (Spline).');
        END;
      ELSE
        FatalError('Zuwenige Vektorelemente (Spline).');
      END;
END Spline;

END Ausgleich.
