IMPLEMENTATION MODULE Fourier;

  (*------------------------------------------------------------------------*)
  (* Modul mit Prozeduren zur Fourieranalyse                                *)
  (*------------------------------------------------------------------------*)
  (* Letzte Aenderung:                                                      *)
  (*                                                                        *)
  (* 13.10.15, MRi: Ersetzen von LFLOAT(#) durch VAL(LONGREAL,#)            *)
  (*------------------------------------------------------------------------*)
  (* Lizenz: GNU General Public License                                     *)
  (*------------------------------------------------------------------------*)

  (* $Id: Fourier.mod,v 1.3 2016/04/15 15:42:33 mriedl Exp mriedl $ *)

FROM LongMath IMPORT pi,sin,cos;
              IMPORT LongComplexMath;

CONST scalarMult = LongComplexMath.scalarMult; 

PROCEDURE four1(VAR data  : ARRAY OF LONGREAL; 
                    nn    : CARDINAL;
                    isign : CARDINAL);

          PROCEDURE sqr(x : LONGREAL) : LONGREAL; BEGIN RETURN x*x; END sqr;

          VAR  ii,jj,n,mmax,m,j,istep,i  : CARDINAL;
               wtemp,wr,wpr,wpi,wi,theta : LONGREAL;
               tempr,tempi,wrs,wis       : LONGREAL;

BEGIN
      n := 2*nn;
      j := 1;
      FOR ii := 1 TO nn DO 
        i := 2*ii-1;
        IF j > i THEN 
          tempr     := data[j-1];
          tempi     := data[j];
          data[j-1] := data[i-1];
          data[j]   := data[i];
          data[i-1] := tempr;
          data[i]   := tempi;
        END;
        m := n DIV 2;
        WHILE (m >= 2) AND (j > m) DO 
          j := j-m;
          m := m DIV 2;
        END;
        j := j+m;
      END; (* FOR ii *)
      mmax := 2;
      WHILE n > mmax DO 
        istep := 2*mmax;
        theta := (2.0*pi) / VAL(LONGREAL,isign*mmax);
        wpr := -2.0*sqr(sin(0.5*theta));
        wpi := sin(theta);
        wr  := 1.0;
        wi  := 0.0;
        FOR ii := 1 TO mmax DIV 2 DO 
          m   := 2*ii-1;
          wrs := wr;
          wis := wi;
          FOR jj := 0 TO (n-m) DIV istep DO 
            i := m + jj*istep;
            j := i + mmax;
            tempr := wrs*data[j-1] - wis*data[j];
            tempi := wrs*data[j]   + wis*data[j-1];
            data[j-1] := data[i-1] - tempr;
            data[j]   := data[i]   - tempi;
            data[i-1] := data[i-1] + tempr;
            data[i]   := data[i]   + tempi;
          END;
          wtemp := wr;
          wr := wr*wpr - wi*wpi    + wr;
          wi := wi*wpr + wtemp*wpi + wi;
        END;
        mmax := istep;
      END;
END four1;

PROCEDURE FourTr(VAR A,B : ARRAY OF LONGREAL;
                 VAR F   : ARRAY OF LONGREAL;
                     n   : CARDINAL);

          VAR  Arg,Ak,Bk,A0,An,rezN,x : LONGREAL;
               k,j                    : CARDINAL;

BEGIN
      rezN := VAL(LONGREAL,n);
      FOR k:=1 TO n-1 DO
        A[k]:=0.0; B[k]:=0.0;
        Arg := pi*VAL(LONGREAL,k) * rezN;
        Ak := 0.0; Bk := 0.0;
        FOR j:=0 TO 2*n-1 DO
          x  := Arg*VAL(LONGREAL,j);
          Ak := Ak + F[j]*cos(x);
          Bk := Bk + F[j]*sin(x);
        END;
        A[k] := Ak * rezN; B[k] := Bk * rezN;
      END;
      A0:=0.0; An:=0.0;
      FOR j:=0 TO 2*n-1 DO
        A0 := A0 + F[j];
        IF ODD(j) THEN An:=An - F[j]; ELSE An:=An + F[j]; END;
      END;
      A[0] := A0*rezN; A[n] := An*rezN;
      B[0] := 0.0;     B[n] := 0.0;
END FourTr;

PROCEDURE Cfft(VAR X    : ARRAY OF LONGCOMPLEX;
                   n    : INTEGER;
                   flag : BOOLEAN);

          VAR  A,N1                : LONGREAL;
               T,W                 : LONGCOMPLEX;
               mr,nn,l,istep,m,i,j : INTEGER;

BEGIN  (* Initialisierung der Variablen *)
      mr := 0;
      nn := n - 1;

      FOR m:=1 TO nn DO (* Bit-Reverse *)
        l := n;
        l := l DIV 2;
        WHILE ((mr + l) > nn) DO l := l DIV 2; END;
        mr := (mr MOD l) + l;
        IF (mr > m) THEN (* Vertauschen *)
          T:=X[m]; X[m]:=X[mr]; X[mr]:=T;
        END;
      END;
 
      l := 1;                 (* fft *)
      WHILE (l < n) DO
        istep := 2 * l;
         FOR m:=1 TO l DO
          A := - pi * VAL(LONGREAL,m - 1) / VAL(LONGREAL,l);
(*
          WITH W DO re:=cos(A); im:=sin(A); 
          IF flag THEN im:= - im; END; END;
 *)
          IF flag THEN
            W := CMPLX(cos(A),-sin(A));
          ELSE
            W := CMPLX(cos(A),+sin(A));
          END;
          i := m - 1;
          REPEAT
            j := i + l;
            T:=W*X[j];
(*
             WITH X[j] DO
              T.re := W.re * re - W.im * im;
              T.im := W.re * im + W.im * re;
            END;
            WITH X[i] DO
              X[j].re := re - T.re;
              X[j].im := im - T.im;
              re := re + T.re;
              im := im + T.im;
            END;
 *)
            X[j]:=X[i] - T;
            X[i]:=X[i] + T;
            INC(i,istep);
          UNTIL (i >= n);
        END;
         l:=istep;
      END;
      IF NOT flag THEN (* Normierung *)
        N1 := 1.0 / VAL(LONGREAL,n);
         FOR i:=0 TO n - 1 DO
          (* WITH X[i] DO re:=re*N1; im:=im*N1; END; *)
          X[i]:=LongComplexMath.scalarMult(N1,X[i]); 
        END;
       END;
END Cfft;

END Fourier.
