IMPLEMENTATION MODULE LinPack;

  (*========================================================================*)
  (* HINWEIS : Bitte nur die Datei LinPack.mod.m2pp editieren               *)
  (*========================================================================*)
  (* Es sind 3 Versionen enthalten die mit                                  *)
  (*                                                                        *)
  (*   m2pp -D __{Parameter}__ < LinPack.mod.m2pp > LibPack.mod             *)
  (*                                                                        *)
  (* mit Parameter = {inline|BLAS|BLASF77} erzeugt werden koennen.          *)
  (*                                                                        *)
  (*   inline : Alle Schleifen werden expliziet ausgefuehrt                 *)
  (*   BLAS   : Schleifen werden durch Routinen aus LibDBlas ausgefuehrt    *)
  (*   BLASF77: Schleifen werden durch Routinen aus der Fortran 77          *)
  (*            Bibiliothek BLAS ausgefuehrt - diese muessen mit gelinkt    *)
  (*            werden - Schnittstelle ist in LibDBlasF77.def beschreiben.  *)
  (*            Zusaetzliche Fortran-Bibiliotheken die gelinkt werden       *)
  (*            muessen sind compilerspezifisch. Level 1 DBLAS mit          *)
  (*            GNU Fortran und ATLAS sind mit XDS M2 getestet.             *)
  (*========================================================================*)
  (* Routines dgefa,dgesl,dgedi and dppfa,dppfi                             *)
  (* from LinPack Fortran 77 libary.                                        *)
  (*                                                                        *)
  (* Licence : GNU public licences                                          *)
  (*                                                                        *)
  (* Last Change:                                                           *)
  (*                                                                        *)
  (* 30.11.15, MRi: Adotion of code form linpack benchmark found in         *)
  (*                Stony Brook Modula-2 examples back to double indexes    *)
  (*                by inspecting provided code and Fortran original source *)
  (* 15.04.16, MRi: Eliminating all address arithmetic which is now done in *)
  (*                LibDBlas. Change from row to colume order.              *)
  (* 17.04.16, MRi: Eliminated final errors ;-)                             *)
  (* 18.04.16, MRi: Introduced preprocessor derectives for inline-code,     *)
  (*                Modula-2 BLAS and Fortran BLAS. Tests look good ...     *)
  (* 21.04.16, MRi: In dgesl FOR k - Schleifen mit BY -1 eingefuegt,        *)
  (*                FOR kb ... DO k=n-nk ... eleminiert                     *)
  (* 22.04.16, MRi: Added procedure dgedi (untested)                        *)
  (* 23.04.16, MRi: inline code in dgedi, error corrected - looks good ;-)  *)
  (* 24.04.16, MRi: Added procedures dppfa and dppdi                        *)
  (* 01.11.17, MRi: Adoped to new underscore-free interface to Fortran BLAS *)
  (*------------------------------------------------------------------------*)
  (* Offene Punkte:                                                         *)
  (*                                                                        *)
  (* - Weiter testen ;-)                                                    *)
  (*------------------------------------------------------------------------*)
  (* Testcodes in TestLinSol.mod and TestPP.mod                             *)
  (*------------------------------------------------------------------------*)
  (* Licence : GNU Lesser General Public License (LGPL)                     *)
  (*------------------------------------------------------------------------*)

  (* $Id: LinPack.mod.m2pp,v 1.10 2016/04/25 08:20:27 mriedl Exp mriedl $ *)

  (*------------------------------------------------------------------------*)
  (* HINT: This version uses Fortran DBLAS level 1 code.                    *)
  (*       FORTRAN 77 DOUBLE PRECISION | Fortran 90 REAL (kind=8) must be   *)
  (*       the same type as Modula-2 LONGREAL                               *)
  (*       It can be linked against ATALS Fortran libralies.                *)
  (*------------------------------------------------------------------------*)

FROM LongMath    IMPORT sqrt;
FROM LibDBlasF77 IMPORT dscal,daxpy,ddot,idamax,dswap;

PROCEDURE dgefa(VAR A    : ARRAY OF ARRAY OF LONGREAL;
                    lda  : INTEGER;
                    n    : INTEGER;
                VAR ipvt : ARRAY OF INTEGER;
                VAR info : INTEGER);

          VAR t             : LONGREAL;
              j,k,kp1,l,nm1 : INTEGER;
              ione,num      : INTEGER;
BEGIN
      (* gaussian elimination with partial pivoting *)

      IF (n > lda) THEN
        (* ... *) 
      END;

      ione:=1; (* for F77 BLAS routines *)
      info := 0;
      nm1 := n - 1;
      IF (nm1 >= 0) THEN
        FOR k:=0 TO nm1-1 DO
          kp1 := k + 1;

          (* find l := pivot index *)

          num:=n-k;
          l := idamax(num,A[k,k],ione) + k - 1; (* -1 wg. F77 *)
          ipvt[k] := l;

          (* zero pivot implies this column already triangularized *)

          IF (A[k,l] = 0.0) THEN
            info := k;
          ELSE

            (* interchange if necessary *)

            IF (l # k) THEN
              t := A[k,l];
              A[k,l] := A[k,k];
              A[k,k] := t;
            END;

            (* compute multipliers *)

            t := -1.0 / A[k,k];
            num:=n-k-1;
            dscal(num,t,A[k,k+1],ione);
            (* row elimination with column indexing *)

            FOR j:=kp1 TO n-1 DO
              t := A[j,l];
              IF (l # k) THEN
                A[j,l] := A[j,k];
                A[j,k] := t;
              END;
              num:=n-k-1;
              daxpy(num,t,A[k,k+1],ione,A[j,k+1],ione);
            END;
          END; (* IF (A[k,l] *)
        END; (* FOR k *)
      END; (* IF nm1 *)
      ipvt[n-1] := n-1;
      IF (A[n-1,n-1] = 0.0) THEN
        info := n-1;
      END;
END dgefa;

PROCEDURE dgesl(VAR A    : ARRAY OF ARRAY OF LONGREAL;
                    lda  : INTEGER;
                    n    : INTEGER;
                VAR ipvt : ARRAY OF INTEGER;
                VAR B    : ARRAY OF LONGREAL;
                    job  : INTEGER);

          (*----------------------------------------------------------------*)
          (* dgesl solves the double precision system                       *)
          (* a * x := b  or  trans(a) * x := b                              *)
          (*                                                                *)
          (* uses daxpy,ddot level 1 blas routine                           *)
          (*----------------------------------------------------------------*)

          VAR t        : LONGREAL;
              k,l,nm1  : INTEGER;
              ione,num : INTEGER;
BEGIN
      IF (n > lda) THEN
        (* ... *)
      END;

      ione:=1; (* for F77 BLAS routines *)

      nm1 := n - 1;
      IF (job # 0) THEN

        (* job = nonzero, solve  trans(a) * x = b  *)
        (* first solve  trans(u)*y = b             *)

        FOR k:=0 TO n-1 DO
          num:=k;
          t := ddot(num,A[k,0],ione,B[0],ione);
          B[k] := (B[k] - t) / A[k,k];
        END;

        (* now solve trans(l)*x = y     *)

        IF (nm1 >= 1) THEN
(*
          FOR kb:=1 TO nm1-1 DO
            k := n - kb - 1;
*)
          FOR k:=nm1-1 TO 0 BY -1 DO (* TO 0 statt TO 1, MRi 26.12.16 *)
            num:=n-k-1;
            B[k]:=B[k] + ddot(num,A[k,k+1],ione,B[k+1],ione);
            l := ipvt[k];
            IF (l # k) THEN
              t    := B[l];
              B[l] := B[k];
              B[k] := t;
            END;
          END;
        END;

      ELSE (* job = 0 , solve  a * x = b   *)

        IF (nm1 >= 1) THEN (* first solve  l*y = b *)
          FOR k:=0 TO nm1-1 DO
            l := ipvt[k];
            t := B[l];
            IF (l # k) THEN
              B[l] := B[k];
              B[k] := t;
            END;
            num:=n-k-1;
            daxpy(num,t,A[k,k+1],ione,B[k+1],ione);
          END;
        END;

        (* now solve  u*x = y *)
(*
        FOR kb:=0 TO n-1 DO
          k := n - 1 - kb;
*)
        FOR k:=n-1 TO 0 BY -1 DO
          B[k] := B[k] / A[k,k];
          t := -B[k];
          num:=k;
          daxpy(num,t,A[k,0],ione,B[0],ione);
        END;
      END;
END dgesl;

PROCEDURE dgedi(VAR A    : ARRAY OF ARRAY OF LONGREAL;
                    lda  : INTEGER;
                    n    : INTEGER;
                VAR ipvt : ARRAY OF INTEGER;
                VAR det  : ARRAY OF LONGREAL; (* [0..2] *)
                VAR work : ARRAY OF LONGREAL;
                    job  : INTEGER);

          VAR i,j,k,l  : INTEGER;
              t        : LONGREAL;
              exit     : BOOLEAN;
              ione,num : INTEGER;
BEGIN
      IF (lda < n) THEN
        (* Fehlerbehandlung *)
      END;
      IF (job DIV 10 # 0) THEN (* Compute the determinant *)
        det[0] := 1.0;
        det[1] := 0.0;
        det[2] := 1.0;
        i:=0; exit:= FALSE;
        WHILE (i <= n-1) AND NOT exit DO (* FOR i:=1 TO n DO *)
          IF (ipvt[i] # i) THEN
            det[0] := -det[0];
            det[2] := -det[2];
          END;
          det[0] := det[0] * A[i,i];
          det[2] := det[2] * A[i,i];
          IF (det[0] = 0.0) THEN
            exit:=TRUE; (* EXIT *)
          ELSE
            WHILE (ABS(det[0]) < 1.0) DO
              det[0] := det[0] * 10.0;
              det[1] := det[1] - 1.0;
            END;
            WHILE (ABS(det[0]) > 10.0) DO
              det[0] := det[0] / 10.0;
              det[1] := det[1] + 1.0;
            END;
            INC(i);
          END;
        END; (* WHILE *)
      END;
  
      (* Compute inverse(U) *)

      ione:=1; (* for F77 BLAS routines *)

      IF ((job MOD 10) # 0) THEN
        FOR k:=0 TO n-1 DO
          A[k,k] := 1.0 / A[k,k];
          t := -A[k,k];
          num:=k;
          dscal(num,t,A[k,0],ione);
          FOR j:=k+1 TO n-1 DO
            t := A[j,k];
            A[j,k] := 0.0;
            num:=k+1;
            daxpy(num,t,A[k,0],ione,A[j,0],ione);
          END;
        END;

        (* Form inverse(U) * inverse(L) *)
  
        FOR k:=n-2 TO 0 BY -1 DO
          FOR i:=k+1 TO n-1 DO
            work[i] := A[k,i];
            A[k,i]  := 0.0;
          END;
          FOR j:=k+1 TO n-1 DO
            t := work[j];
            num:=n;
            daxpy(num,t,A[j,0],ione,A[k,0],ione);
          END;
          l := ipvt[k];
          IF (l # k) THEN
            num:=n;
            dswap(num,A[k,0],ione,A[l,0],ione);
          END;
        END;
      END;
END dgedi;

PROCEDURE dppfa(VAR AP   : ARRAY OF LONGREAL; (* SUPERVEKTOR *)
                    n    : INTEGER;
                VAR Info : INTEGER);

          (* factor a symmetric positive definite matrix *)

          VAR j,jj,k,kj,kk : INTEGER;
              s,t          : LONGREAL;
              num,ione     : INTEGER;
BEGIN 
      Info:=0;

      ione:=1; (* Wg. F77 Blas *)

      jj:=0;
      FOR j:=1 TO n DO
        s := 0.0;
        kj := jj;
        kk := 0;
        FOR k:=1 TO j-1 DO
          num:=k-1;
          t := AP[kj] - ddot(num,AP[kk],ione,AP[jj],ione);
          INC(kk,k);
          t := t / AP[kk-1];
          AP[kj] := t;
          s:=s + t*t;
          INC(kj);
        END;
        INC(jj,j);
        s := AP[jj-1] - s;
        IF (s <= 0.0) THEN
          Info:=j;
          RETURN;
        END;
        AP[jj-1] := sqrt(s);
      END;
END dppfa;

PROCEDURE dppdi(VAR AP  : ARRAY OF LONGREAL; (* SUPERVEKTOR *)
                    n   : INTEGER;
                VAR det : ARRAY OF LONGREAL; (* 2 Elemente *)
                    job : INTEGER);

          VAR i,ii,j,j1,jj,k,k1,kj,kk : INTEGER;
              t                       : LONGREAL;
              exit                    : BOOLEAN;
              num,ione                : INTEGER;
BEGIN
      IF (job DIV 10 # 0) THEN (* Compute the determinant *)
        det[0] := 1.0;
        det[1] := 0.0;
        det[2] := 1.0;
        ii := 0;
        i:=1; exit:= FALSE;
        WHILE (i <= n) AND NOT exit DO (* FOR i:=1 TO n DO *)
          INC(ii,i);
          det[0] := det[0] * AP[ii-1]*AP[ii-1];
          det[2] := det[2] * AP[ii-1]*AP[ii-1];
          IF (det[0] = 0.0) THEN
            exit:=TRUE; (* EXIT *)
          ELSE
            WHILE (det[0] < 1.0) DO
              det[0] := det[0] * 10.0;
              det[1] := det[1] - 1.0;
            END;
            WHILE (det[0] > 10.0) DO
              det[0] := det[0] / 10.0;
              det[1] := det[1] + 1.0;
            END;
            INC(i);
          END;
        END; (* WHILE *)
      END;
  
      (* Compute inverse(R) *)
  
      IF ((job MOD 10) # 0) THEN

        ione:=1; (* Wg. F77 Blas *)

        kk := 0;
        FOR k:=1 TO n DO
          k1 := kk;
          INC(kk,k);
          AP[kk-1] := 1.0 / AP[kk-1];
          t := -AP[kk-1];
          num := k-1;
          dscal(num,t,AP[k1],ione);
          j1 := kk;
          kj := kk + k - 1;
          FOR j:=k+1 TO n DO
            t      := AP[kj];
            AP[kj] := 0.0;
            num := k;
            daxpy(num,t,AP[k1],ione,AP[j1],ione);
            INC(j1,j);
            INC(kj,j);
          END;
        END;
  
        (* Form inverse(R) * (inverse(R))' *)
  
        jj := 0;
        FOR j:=1 TO n DO
          j1 := jj;
          kj := jj;
          k1 := 0;
          FOR k:=1 TO j-1 DO
            t := AP[kj];
            num := k;
            daxpy(num,t,AP[j1],ione,AP[k1],ione);
            INC(k1,k);
            INC(kj);
          END;
          INC(jj,j);
          t := AP[jj-1];
          num:=j;
          dscal(num,t,AP[j1],ione);
        END;
      END;
END dppdi;

END LinPack.
