IMPLEMENTATION MODULE SpezFunkt3;

  (*------------------------------------------------------------------------*)
  (* Letzte Bearbeitung:                                                    *)
  (*                                                                        *)
  (* 14.05.16, MRi: Prozeduren Bessel, Y01, BessJ0, H01 aus LMathLib her-   *)
  (*                ausgenommen und Modul SpezFunkt3 erstellt               *)
  (* 29.05.16, MRi: Bessel in BesselJN, BessJ0 in BesselJ0, Y01 in          *)
  (*                BesselY0 umbenannt.                                     *)
  (* 30.05.16, MRi: H01 in Hankel01 umbenannt, Fehler korrigiert            *)
  (*------------------------------------------------------------------------*)
  (* Testroutinen TestSpFun3.mod & TestSpFun3.f (Fortran)                   *)
  (*------------------------------------------------------------------------*)
  (* Licence : GNU Lesser General Public License (LGPL)                     *)
  (*------------------------------------------------------------------------*)

  (*
   * $Id: SpezFunkt3.mod,v 1.4 2016/05/30 08:48:09 mriedl Exp mriedl $
   *)

FROM LongMath IMPORT sqrt,sin,cos,ln,pi;
FROM Errors   IMPORT ErrOut;
              IMPORT LMathLib;

PROCEDURE BesselJN(x : LONGREAL;
                   n : CARDINAL) : LONGREAL;

          VAR  xh2ipn,xh,Z,Sum,Fakt,Faktipn  : LONGREAL;
               i,ipn                         : CARDINAL;
BEGIN
      xh:=0.5*x;
      xh2ipn:=LMathLib.CardPot(xh,n);
      xh:=xh*xh;
      Sum:=0.0;  i:=0; ipn:=n; (* ipn=i+n *)
      Fakt:=1.0; Faktipn:=LMathLib.fact(n);
      LOOP
        Z:=Sum;
        IF ODD(i) THEN
          Sum:=Sum - xh2ipn/(Fakt*Faktipn);
        ELSE
          Sum:=Sum + xh2ipn/(Fakt*Faktipn);
        END;
        IF (Z = Sum) THEN EXIT END;
        INC(i); INC(ipn);
        Fakt:=Fakt*LFLOAT(i);
        Faktipn:=Faktipn*LFLOAT(ipn);
        xh2ipn:=xh2ipn*xh;
      END; (* LOOP *)
      RETURN Sum;
END BesselJN;

PROCEDURE BesselJ0(x : LONGREAL) : LONGREAL;

          VAR  ax,xx,z         : LONGREAL;
               y,ans,ans1,ans2 : LONGREAL;

          (* Numerical recipes - austauschen !!! *)

BEGIN
      IF (ABS(x) < 8.0) THEN 
        y := x*x;
        ans1 := 57568490574.0 + 
                y*(-13362590354.0 + y*(651619640.7 +
                y*(-11214424.18   + y*(77392.33017 +
                y*(-184.9052456)))));
        ans2 := 57568490411.0 +
                y*(1029532985.0 + y*(9494680.718 +
                y*(59272.64853  + y*(267.8532712 + y))));
        RETURN ans1 / ans2;
      ELSE 
        ax := ABS(x);
        z := 8.0 / ax;
        y := z*z;
        xx := ax - 0.785398164;
        ans1 := 1.0 +
                y*(-0.1098628627E-02 + y*(0.2734510407E-04 +
                y*(-0.2073370639E-05 + y*0.2093887211E-06)));
        ans2 := -0.1562499995E-01 + 
                y*(0.1430488765E-03 + y*(-0.6911147651E-05 +
                y*(0.7621095161E-06 - y*0.934945152E-07)));
        ans := sqrt(0.636619772 / ax) * (cos(xx)*ans1 - z*sin(xx)*ans2);
        RETURN ans;
      END;
END BesselJ0;

PROCEDURE BesselY0(x : LONGREAL) : LONGREAL;

      CONST  C  = 0.577215664901532; (* Eulerkonstante. *)

      VAR    xh2,y,Z,Sum,Fakt,Phi : LONGREAL;
             i                    : CARDINAL;

BEGIN
      xh2:=0.25*x*x; Sum:=-2.0*xh2;
      i:=2;          Fakt:=2.0;
      Phi:=1.5;      y:=xh2*xh2;
      LOOP
        Z:=Sum;
        IF ODD(i) THEN
          Sum:=Sum - 2.0*y*Phi/(Fakt*Fakt);
        ELSE
          Sum:=Sum + 2.0*y*Phi/(Fakt*Fakt);
        END; (* IF *)
        IF (Z = Sum) THEN EXIT END;
        INC(i);
        Fakt:=Fakt*LFLOAT(i);
        y:=y*xh2;
        Phi:=Phi + 1.0 / LFLOAT(i);
      END; (* LOOP *)
      RETURN (2.0*(C + ln(x/2.0))*BesselJN(x,0) - Sum) / pi;
END BesselY0;

PROCEDURE Hankel01(x : LONGREAL) : LONGCOMPLEX;

          CONST C  = LMathLib.C; (* Eulerkonstante *)
          
          VAR   xh,xh2,Zi,SumR,SumI,Phi,yf2,Fakt,y : LONGREAL;
                i                                  : CARDINAL;
BEGIN                
      IF (x <= 0.0) THEN
        ErrOut('x = 0.0 (SpezFunkt3.Hankel01)');
        RETURN CMPLX(0.0,0.0);
      END;
      xh:=0.5*x; xh2:=xh*xh;
      SumR:=1.0; SumI:=0.0;
      Fakt:=1.0; Phi:=1.0;
      i:=1;      y:=xh2;
      LOOP
        Zi  := SumI;
        yf2 := y / (Fakt*Fakt);
        IF ODD(i) THEN
          SumR:=SumR - yf2;
          SumI:=SumI - 2.0*yf2*Phi;
        ELSE
          SumR:=SumR + yf2;
          SumI:=SumI + 2.0*yf2*Phi;
        END;
        IF (Zi = SumI) THEN EXIT END; (* !!!! *)
        INC(i);
        Fakt:=Fakt*VAL(LONGREAL,i);
        Phi:=Phi + 1.0/VAL(LONGREAL,i);
        y:=y*xh2;
      END; (* LOOP *)
      RETURN CMPLX(SumR , ((2.0*SumR*(C + ln(xh)) - SumI) / pi));
END Hankel01;

END SpezFunkt3.
